@@ -190,6 +190,7 @@ sub compile
     my $self = shift;
     my %p = validate( @_, { comp_source => { type => SCALAR|SCALARREF },
                             name => { type => SCALAR },
+                            comp_path => { type => SCALAR },
                             fh => { type => HANDLE, optional => 1 },
                           } );
     my $src = ref($p{comp_source}) ? $p{comp_source} : \$p{comp_source};
@@ -203,6 +204,8 @@ sub compile
     # So we're re-entrant in subcomps
     local $self->{paused_compiles} = [];
 
+    local $self->{comp_path} = $p{comp_path};
+
     # Preprocess the source.  The preprocessor routine is handed a
     # reference to the entire source.
     if ($self->preprocess)
@@ -222,7 +225,6 @@ sub start_component
     my $c = $self->{current_compile};
 
     $c->{in_main} = 1;
-    $c->{comp_with_content_stack} = [];
 
     $c->{in_block} = undef;
 
@@ -246,6 +248,8 @@ sub _init_comp_data
     $data->{flags} = {};
     $data->{attr} = {};
 
+    $data->{comp_with_content_stack} = [];
+
     foreach ( qw( cleanup filter init once shared ) )
     {
         $data->{blocks}{$_} = [];
@@ -258,7 +262,7 @@ sub end_component
     my $c = $self->{current_compile};
 
     $self->lexer->throw_syntax_error("Not enough component-with-content ending tags found")
-        if $c->{comp_with_content_stack} && @{ $c->{comp_with_content_stack} };
+        if @{ $c->{comp_with_content_stack} };
 }
 
 sub start_block
@@ -812,9 +816,33 @@ a Compiler subclass needs to implement.
 
 =over 4
 
-=item compile(comp_source => <string>, name => <string>, comp_class => <string>)
+=item compile(...)
 
-The "comp_class" parameter may be ignored by the compiler.
+This method has several parameters:
+
+=over 8
+
+=item * comp_source (required)
+
+Either a scalar or reference to a scalar containing the component source.
+
+=item * name (required)
+
+The name of the component. This should be the filename of the
+component if it is file-based, or some other clear identifier of the
+component source.
+
+=item * comp_path (required)
+
+This should be the component's path.
+
+=item * fh (optional)
+
+If this is given then the output of the compiler will be sent directly
+to this handle, rather than being buffered in memory. This is an
+optimization to avoid memory usage.
+
+=back
 
 =item object_id
 
