@@ -1,13 +1,14 @@
-# Copyright (c) 1998-2003 by Jonathan Swartz. All rights reserved.
+# Copyright (c) 1998-2005 by Jonathan Swartz. All rights reserved.
 # This program is free software; you can redistribute it and/or modify
 # it under the same terms as Perl itself.
 
 package HTML::Mason::Compiler::ToObject;
 
 use strict;
+use warnings;
 
-use Params::Validate qw(SCALAR validate);
-use HTML::Mason::Tools qw(make_fh taint_is_on);
+use Params::Validate qw(BOOLEAN SCALAR validate);
+use HTML::Mason::Tools qw(taint_is_on);
 
 use HTML::Mason::Compiler;
 use base qw( HTML::Mason::Compiler );
@@ -20,28 +21,28 @@ use File::Basename qw(dirname);
 BEGIN
 {
     __PACKAGE__->valid_params
-	(
-	 comp_class =>
+        (
+         comp_class =>
          { parse => 'string', type => SCALAR, default => 'HTML::Mason::Component',
            descr => "The class into which component objects will be blessed" },
 
-	 subcomp_class =>
+         subcomp_class =>
          { parse => 'string', type => SCALAR, default => 'HTML::Mason::Component::Subcomponent',
            descr => "The class into which subcomponent objects will be blessed" },
 
-	 in_package =>
+         in_package =>
          { parse => 'string', type => SCALAR, default => 'HTML::Mason::Commands',
            descr => "The package in which component execution will take place" },
 
-	 preamble =>
+         preamble =>
          { parse => 'string', type => SCALAR, default => '',
            descr => "A chunk of Perl code to add to the beginning of each compiled component" },
 
-	 postamble =>
+         postamble =>
          { parse => 'string', type => SCALAR, default => '',
            descr => "A chunk of Perl code to add to the end of each compiled component" },
 
-	 use_strict =>
+         use_strict =>
          { parse => 'boolean', type => SCALAR, default => 1,
            descr => "Whether to turn on Perl's 'strict' pragma in components" },
 
@@ -49,22 +50,26 @@ BEGIN
          { parse => 'string', type => SCALAR, default => 'auto',
            regex => qr/^(?:always|auto|never)$/,
            descr => "Whether or not to create the %ARGS hash" },
-	);
+
+         named_component_subs =>
+         { parse => 'boolean', type => BOOLEAN, default => 0,
+           descr => "Whether to use named subroutines for component code" },
+        );
 }
 
 use HTML::Mason::MethodMaker
-    ( read_write => [ map { [ $_ => __PACKAGE__->validation_spec->{$_} ] }
-		      qw( comp_class
-                          define_args_hash
-                          in_package
-			  postamble
-			  preamble
-                          subcomp_class
-			  use_strict
+    ( read_only => [
+                    qw( comp_class
+                        define_args_hash
+                        in_package
+                        named_component_subs
+                        postamble
+                        preamble
+                        subcomp_class
+                        use_strict
                         )
-		    ],
-    );
-
+                    ],
+      );
 
 sub compile
 {
@@ -89,59 +94,50 @@ sub compile_to_file
     my $self = shift;
 
     my %p = validate( @_, {   file => { type => SCALAR },
-			    source => { isa => 'HTML::Mason::ComponentSource' } },
-		    );
+                            source => { isa => 'HTML::Mason::ComponentSource' } },
+                    );
 
     my ($file, $source) = @p{qw(file source)};
     my @newfiles = ($file);
 
     if (defined $file && !-f $file) {
-	my ($dirname) = dirname($file);
-	if (!-d $dirname) {
-	    unlink($dirname) if (-e _);
-	    push @newfiles, mkpath($dirname, 0, 0775);
-	    system_error "Couldn't create directory $dirname: $!"
-		unless -d $dirname;
-	}
-	rmtree($file) if (-d $file);
+        my ($dirname) = dirname($file);
+        if (!-d $dirname) {
+            unlink($dirname) if (-e _);
+            push @newfiles, mkpath($dirname, 0, 0775);
+            system_error "Couldn't create directory $dirname: $!"
+                unless -d $dirname;
+        }
+        rmtree($file) if (-d $file);
     }
 
     ($file) = $file =~ /^(.*)/s if taint_is_on;  # Untaint blindly
 
-    my $fh = make_fh();
-    open $fh, "> $file"
-	or system_error "Couldn't create object file $file: $!";
+    open my $fh, "> $file"
+        or system_error "Couldn't create object file $file: $!";
 
     $self->compile( comp_source => $source->comp_source_ref,
-		    name => $source->friendly_name,
-		    comp_class => $source->comp_class,
-		    fh => $fh );
+                    name => $source->friendly_name,
+                    comp_class => $source->comp_class,
+                    comp_path => $source->comp_path,
+                    fh => $fh );
 
     close $fh 
-	or system_error "Couldn't close object file $file: $!";
+        or system_error "Couldn't close object file $file: $!";
     
     return \@newfiles;
 }
 
-sub object_id
-{
-    my $self = shift;
-
-    local $self->{comp_class} = '';
-
-    return $self->SUPER::object_id;
-}
-
 sub _output_chunk
 {
     my ($self, $fh, $string) = (shift, shift, shift);
     if ($fh)
     {
-	print $fh (ref $_ ? $$_ : $_) foreach grep defined, @_;
+        print $fh (ref $_ ? $$_ : $_) foreach grep defined, @_;
     }
     else
     {
-	$$string .= (ref $_ ? $$_ : $_) foreach @_;
+        $$string .= (ref $_ ? $$_ : $_) foreach @_;
     }
 }
 
@@ -157,19 +153,12 @@ sub compiled_component
     local $c->{compiled_def} = $self->_compile_subcomponents if %{ $c->{def} };
     local $c->{compiled_method} = $self->_compile_methods if %{ $c->{method} };
 
-    # Create the file header to assert creatorship
-    my $id = $self->object_id;
-    $id =~ s,([\\']),\\$1,g;
-    $self->_output_chunk($p{fh}, \$obj_text, "# MASON COMPILER ID: $id\n");
-
     # Some preamble stuff, including 'use strict', 'use vars', and <%once> block
     my $header = $self->_make_main_header;
     $self->_output_chunk($p{fh}, \$obj_text, $header);
 
-
     my $params = $self->_component_params;
 
-    $params->{compiler_id} = "'$id'";
     $params->{load_time} = time;
 
     $params->{subcomps} = '\%_def' if %{ $c->{def} };
@@ -177,69 +166,100 @@ sub compiled_component
 
     if ( $self->_blocks('shared') )
     {
-	my %subs;
-	while ( my ($name, $pref) = each %{ $c->{compiled_def} } )
-	{
-	    my $key = "subcomponent_$name";
-	    $subs{$key} = $pref->{code};
-	    $pref->{code} = "sub {\nHTML::Mason::Request->instance->call_dynamic( '$key', \@_ )\n}";
-	}
-	while (my ($name, $pref) = each %{ $c->{compiled_method} } )
-	{
-	    my $key = "method_$name";
-	    $subs{$key} = $pref->{code};
-	    $pref->{code} = "sub {\nHTML::Mason::Request->instance->call_dynamic( '$key', \@_ )\n}";
-	}
-	$subs{main} = $params->{code};
-	$params->{code} = "sub {\nHTML::Mason::Request->instance->call_dynamic( 'main', \@_ )\n}";
+        my %subs;
+        while ( my ($name, $pref) = each %{ $c->{compiled_def} } )
+        {
+            my $key = "subcomponent_$name";
+            $subs{$key} = $pref->{code};
+            $pref->{code} = "sub {\nHTML::Mason::Request->instance->call_dynamic('$key',\@_)\n}";
+        }
+        while (my ($name, $pref) = each %{ $c->{compiled_method} } )
+        {
+            my $key = "method_$name";
+            $subs{$key} = $pref->{code};
+            $pref->{code} = "sub {\nHTML::Mason::Request->instance->call_dynamic( '$key', \@_ )\n}";
+        }
+        $subs{main} = $params->{code};
+        $params->{code} = "sub {\nHTML::Mason::Request->instance->call_dynamic( 'main', \@_ )\n}";
 
-	$params->{dynamic_subs_init} =
-	    join '', ( "sub {\n",
-		       $self->_blocks('shared'),
-		       "return {\n",
-		       map( "'$_' => $subs{$_},\n", sort keys %subs ),
-		       "\n}\n}"
-		     );
+        my $named_subs = '';
+        my %named_subs = $self->_named_subs_hash;
+        while ( my ( $name, $body ) = each %named_subs )
+        {
+            $named_subs .= '*' . $name . " = sub {\n" . $body . "\n};\n\n";
+        }
+
+        $params->{dynamic_subs_init} =
+            join '', ( "sub {\n",
+                       $self->_set_request,
+                       $self->_blocks('shared'),
+                       $named_subs,
+                       "return {\n",
+                       map( "'$_' => $subs{$_},\n", sort keys %subs ),
+                       "\n}\n}"
+                     );
+    }
+    else
+    {
+        my %named_subs = $self->_named_subs_hash;
+        while ( my ( $name, $body ) = each %named_subs )
+        {
+            $self->_output_chunk( $p{fh}, \$obj_text,
+                                  "sub $name {\n" . $body . "\n}\n"
+                                );
+        }
     }
 
     $self->_output_chunk($p{fh}, \$obj_text, $self->_subcomponents_footer);
     $self->_output_chunk($p{fh}, \$obj_text, $self->_methods_footer);
 
-
-
-    $params->{object_size} = 0;
-    $params->{object_size} += length for ($header, %$params);
-
     $self->_output_chunk($p{fh}, \$obj_text,
-			 $self->_constructor( $self->comp_class,
-					      $params ),
-			 ';',
-			);
+                         $self->_constructor( $self->comp_class,
+                                              $params ),
+                         ';',
+                        );
 
     return \$obj_text;
 }
 
-sub assert_creatorship
+sub _named_subs_hash
 {
-    my ($self, $p) = @_;
-    my $id;
-    if ($p->{object_code}) {
-	# Read the object code as a string
+    my $self = shift;
 
-	($id) = ${$p->{object_code}} =~ /\A# MASON COMPILER ID: (\S+)$/m
-	    or wrong_compiler_error "Couldn't find a Compiler ID in compiled code.";
-    } else {
-	# Open the object file and read its first line
+    return unless $self->named_component_subs;
 
-	my $fh = make_fh();
-	open $fh, $p->{object_file} or die "Can't read $p->{object_file}: $!";
-	($id) = <$fh> =~ /\A# MASON COMPILER ID: (\S+)$/m
-	    or wrong_compiler_error "Couldn't find a Compiler ID in $p->{object_file}.";
-	close $fh;
+    my %subs;
+    $subs{ $self->_sub_name } = $self->_body;
+
+    while ( my ( $name, $params ) =
+            each %{ $self->{current_compile}{compiled_def} } )
+    {
+        $subs{ $self->_sub_name( 'def', $name ) } = $params->{body};
     }
-    
-    wrong_compiler_error 'This object file was created by an incompatible Compiler or Lexer.  Please remove the component files in your object directory.'
-	unless $id eq $self->object_id;
+
+    while ( my ( $name, $params ) =
+            each %{ $self->{current_compile}{compiled_method} } )
+    {
+        $subs{ $self->_sub_name( 'method', $name ) } = $params->{body};
+    }
+
+    return %subs;
+}
+
+sub _sub_name
+{
+    my $self = shift;
+
+    return join '_', $self->_escape_sub_name_part( $self->{comp_path}, @_ );
+}
+
+sub _escape_sub_name_part
+{
+    my $self = shift;
+
+    return map { my $part = $_;
+                 $part =~ s/([^\w_])/'_' . sprintf( '%x', ord $1 )/ge;
+                 $part; } @_;
 }
 
 sub _compile_subcomponents
@@ -264,9 +284,9 @@ sub _compile_subcomponents_or_methods
     my %compiled;
     foreach ( keys %{ $self->{current_compile}{$type} } )
     {
-	local $self->{current_compile} = $self->{current_compile}{$type}{$_};
-	local $self->{current_compile}->{in_named_block} = {type => $type, name => $_};
-	$compiled{$_} = $self->_component_params;
+        local $self->{current_compile} = $self->{current_compile}{$type}{$_};
+        local $self->{current_compile}->{in_named_block} = {type => $type, name => $_};
+        $compiled{$_} = $self->_component_params;
     }
 
     return \%compiled;
@@ -279,11 +299,11 @@ sub _make_main_header
     my $pkg = $self->in_package;
 
     return join '', ( "package $pkg;\n",
-		      $self->use_strict ? "use strict;\n" : "no strict;\n",
-		      sprintf( "use vars qw(\%s);\n",
-			       join ' ', '$m', $self->allow_globals ),
-		      $self->_blocks('once'),
-		    );
+                      $self->use_strict ? "use strict;\n" : "no strict;\n",
+                      sprintf( "use vars qw(\%s);\n",
+                               join ' ', '$m', $self->allow_globals ),
+                      $self->_blocks('once'),
+                    );
 }
 
 sub _subcomponents_footer
@@ -309,13 +329,13 @@ sub _subcomponent_or_method_footer
     return '' unless %{ $c->{$type} };
 
     return join('',
-		"my %_$type =\n(\n",
-		map( {("'$_' => " ,
-		       $self->_constructor( $self->{subcomp_class},
-					    $c->{"compiled_$type"}{$_} ) ,
-		       ",\n")} keys %{ $c->{"compiled_$type"} } ) ,
-		"\n);\n"
-	       );
+                "my %_$type =\n(\n",
+                map( {("'$_' => " ,
+                       $self->_constructor( $self->{subcomp_class},
+                                            $c->{"compiled_$type"}{$_} ) ,
+                       ",\n")} keys %{ $c->{"compiled_$type"} } ) ,
+                "\n);\n"
+               );
 }
 
 sub _constructor
@@ -323,17 +343,32 @@ sub _constructor
     my ($self, $class, $params) = @_;
 
     return ("${class}->new(\n",
-	    map( {("'$_' => ", $params->{$_}, ",\n")} sort keys %$params ),
-	    "\n)\n",
-	   );
+            map( {("'$_' => ", $params->{$_}, ",\n")}
+                 sort grep { $_ ne 'body' } keys %$params ),
+            "\n)\n",
+           );
 }
 
 sub _component_params
 {
     my $self = shift;
 
-    my %params = ( code => join ( '', "sub {\n", $self->_body, "}" ),
-		 );
+    my %params;
+
+    if ( $self->named_component_subs )
+    {
+        $params{code} =
+            '\\&' .
+            $self->_sub_name
+                ( grep { defined }
+                  @{ $self->{current_compile}{in_named_block} }
+                  { 'type', 'name' } );
+        $params{body} = $self->_body;
+    }
+    else
+    {
+        $params{code} = join '', "sub {\n", $self->_body, "}";
+    }
 
     $params{flags} = join '', "{\n", $self->_flags, "\n}"
         if keys %{ $self->{current_compile}{flags} };
@@ -342,7 +377,7 @@ sub _component_params
         if keys %{ $self->{current_compile}{attr} };
 
     $params{declared_args} = join '', "{\n", $self->_declared_args, "\n}"
-	if @{ $self->{current_compile}{args} };
+        if @{ $self->{current_compile}{args} };
 
     $params{has_filter} = 1 if $self->_blocks('filter');
 
@@ -355,15 +390,25 @@ sub _body
 
     return join '', ( $self->preamble,
                       $self->_set_request,
-		      $self->_arg_declarations,
+                      $self->_set_buffer,
+                      $self->_arg_declarations,
                       $self->_filter,
-		      "\$m->debug_hook( \$m->current_comp->path ) if ( \%DB:: );\n\n",
-		      $self->_blocks('init'),
-		      $self->{current_compile}{body},
-		      $self->_blocks('cleanup'),
-		      $self->postamble,
-		      "return undef;\n",
-		    );
+                      "\$m->debug_hook( \$m->current_comp->path ) if ( HTML::Mason::Compiler::IN_PERL_DB() );\n\n",
+                      $self->_blocks('init'),
+
+                      # do not add a block around this, it introduces
+                      # a separate scope and might break cleanup
+                      # blocks (or all sort of other things!)
+                      $self->{current_compile}{body},
+
+                      $self->_blocks('cleanup'),
+                      $self->postamble,
+
+                      # semi before return will help catch syntax
+                      # errors in component body - don't return values
+                      # explicitly
+                      ";return;\n",
+                    );
 }
 
 sub _set_request
@@ -375,9 +420,20 @@ sub _set_request
     return 'local $' . $self->in_package . '::m = $HTML::Mason::Commands::m;' . "\n";
 }
 
+sub _set_buffer
+{
+    my $self = shift;
+
+    if ($self->enable_autoflush) {
+        return '';
+    } else {
+        return 'my $_outbuf = $m->{top_stack}->[HTML::Mason::Request::STACK_BUFFER];' . "\n";
+    }
+}
+
 my %coercion_funcs = ( '@' => 'HTML::Mason::Tools::coerce_to_array',
-		       '%' => 'HTML::Mason::Tools::coerce_to_hash',
-		     );
+                       '%' => 'HTML::Mason::Tools::coerce_to_hash',
+                     );
 sub _arg_declarations
 {
     my $self = shift;
@@ -446,39 +502,39 @@ EOF
 
     foreach ( @{ $self->{current_compile}{args} } )
     {
-	my $var_name = "$_->{type}$_->{name}";
-	push @decl, $var_name;
+        my $var_name = "$_->{type}$_->{name}";
+        push @decl, $var_name;
 
         my $arg_in_array = "\$_[ \$pos{'$_->{name}'} ]";
 
-	my $coerce;
-	if ( $coercion_funcs{ $_->{type} } )
-	{
-	    $coerce = $coercion_funcs{ $_->{type} } . "( $arg_in_array, '$var_name')";
-	}
-	else
-	{
-	    $coerce = $arg_in_array;
-	}
+        my $coerce;
+        if ( $coercion_funcs{ $_->{type} } )
+        {
+            $coerce = $coercion_funcs{ $_->{type} } . "( $arg_in_array, '$var_name')";
+        }
+        else
+        {
+            $coerce = $arg_in_array;
+        }
 
-	push @assign, "#line $_->{line} $_->{file}\n"
-	    if defined $_->{line} && defined $_->{file} && $self->use_source_line_numbers;
+        push @assign, "#line $_->{line} $_->{file}\n"
+            if defined $_->{line} && defined $_->{file} && $self->use_source_line_numbers;
 
-	if ( defined $_->{default} )
-	{
-	    my $default_val = $_->{default};
-	    # allow for comments after default declaration
-	    $default_val .= "\n" if defined $_->{default} && $_->{default} =~ /\#/;
+        if ( defined $_->{default} )
+        {
+            my $default_val = $_->{default};
+            # allow for comments after default declaration
+            $default_val .= "\n" if defined $_->{default} && $_->{default} =~ /\#/;
 
-	    push @assign, <<"EOF";
+            push @assign, <<"EOF";
      $var_name = exists \$pos{'$_->{name}'} ? $coerce : $default_val;
 EOF
-	}
-	else
-	{
-	    push @assign,
-		"    $var_name = $coerce;\n";
-	}
+        }
+        else
+        {
+            push @assign,
+                "    $var_name = $coerce;\n";
+        }
     }
 
     my $decl = 'my ( ';
@@ -547,7 +603,7 @@ sub _flags_or_attr
     my $type = shift;
 
     return join "\n,", ( map { "$_ => $self->{current_compile}{$type}{$_}" }
-			 keys %{ $self->{current_compile}{$type} } );
+                         keys %{ $self->{current_compile}{$type} } );
 }
 
 sub _declared_args
@@ -557,13 +613,13 @@ sub _declared_args
     my @args;
 
     foreach my $arg ( sort {"$a->{type}$a->{name}" cmp "$b->{type}$b->{name}" }
-		      @{ $self->{current_compile}{args} } )
+                      @{ $self->{current_compile}{args} } )
     {
-	my $def = defined $arg->{default} ? "$arg->{default}" : 'undef';
-	$def =~ s,([\\']),\\$1,g;
-	$def = "'$def'" unless $def eq 'undef';
+        my $def = defined $arg->{default} ? "$arg->{default}" : 'undef';
+        $def =~ s,([\\']),\\$1,g;
+        $def = "'$def'" unless $def eq 'undef';
 
-	push @args, "  '$arg->{type}$arg->{name}' => { default => $def }";
+        push @args, "  '$arg->{type}$arg->{name}' => { default => $def }";
     }
 
     return join ",\n", @args;
@@ -581,7 +637,11 @@ HTML::Mason::Compiler::ToObject - A Compiler subclass that generates Mason objec
 
   my $compiler = HTML::Mason::Compiler::ToObject->new;
 
-  my $object_code = $compiler->compile( comp_source => $source, name => $comp_name );
+  my $object_code =
+      $compiler->compile( comp_source => $source,
+                          name        => $comp_name,
+                          comp_path   => $comp_path,
+                        );
 
 =head1 DESCRIPTION
 
@@ -611,17 +671,29 @@ historical reasons, this defaults to C<HTML::Mason::Commands>.
 
 =item preamble
 
-Text given for this parameter is placed at the beginning of each component. See also L<postamble|HTML::Mason::Params/postamble>.
+Text given for this parameter is placed at the beginning of each
+component, but after the execution of any C<< <%once> >> block. See
+also L<postamble|HTML::Mason::Params/postamble>. The request will be available as C<$m> in preamble
+code.
 
 =item postamble
 
-Text given for this parameter is placed at the end of each component. See also L<preamble|HTML::Mason::Params/preamble>.
+Text given for this parameter is placed at the end of each
+component. See also L<preamble|HTML::Mason::Params/preamble>.  The request will be available as
+C<$m> in postamble code.
 
 =item use_strict
 
 True or false, default is true. Indicates whether or not a given
 component should C<use strict>.
 
+=item named_component_subs
+
+When compiling a component, use uniquely named subroutines for the a
+component's body, subcomponents, and methods. Doing this allows you to
+effectively profile Mason components. Without this, all components
+simply show up as __ANON__ or something similar in the profiler.
+
 =item define_args_hash
 
 One of "always", "auto", or "never".  This determines whether or not
@@ -638,6 +710,13 @@ arguments, which can save memory and slightly improve execution speed.
 
 =back
 
+=head1 ACCESSOR METHODS
+
+All of the above properties have read-only accessor methods of the
+same name. You cannot change any property of a compiler after it has
+been created (but you can create multiple compilers with different
+properties).
+
 =head1 METHODS
 
 This class is primarily meant to be used by the Interpreter object,
@@ -645,12 +724,14 @@ and as such has a very limited public API.
 
 =over
 
-=item compile (comp_source => $source, name => $name, comp_class = $comp_class)
+=item compile(...)
 
 This method will take component source and return the compiled object
-code for that source.  The C<comp_source> and C<name> parameters are
-optional.  The C<comp_class> can be used to change the component class
-for this one comonent.
+code for that source. See L<HTML::Mason::Compiler/compile(...)> for
+details on this method.
+
+This subclass also accepts a C<comp_class> parameter, allowing you to
+override the class into which the component is compiled.
 
 =back
 
